@@ -1,8 +1,6 @@
 {-# LANGUAGE TypeSynonymInstances, FlexibleInstances, MultiParamTypeClasses,
-    DeriveDataTypeable, GeneralizedNewtypeDeriving, CPP, StandaloneDeriving #-}
-#ifdef GENERICS
-{-# LANGUAGE DeriveGeneric #-}
-#endif
+    DeriveDataTypeable, GeneralizedNewtypeDeriving, CPP, StandaloneDeriving,
+    DeriveGeneric #-}
 {-
 Copyright (C) 2010-2012 John MacFarlane <jgm@berkeley.edu>
 
@@ -32,9 +30,8 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 Convenience functions for building pandoc documents programmatically.
 
-Example of use:
+Example of use (with @OverloadedStrings@ pragma):
 
-> {-# LANGUAGE OverloadedStrings #-}
 > import Text.Pandoc.Builder
 >
 > myDoc :: Pandoc
@@ -49,18 +46,19 @@ Example of use:
 Isn't that nicer than writing the following?
 
 > import Text.Pandoc.Definition
+> import Data.Map (fromList)
 >
 > myDoc :: Pandoc
-> myDoc = Pandoc (Meta {docTitle = [Str "My",Space,Str "title"]
->                      , docAuthors = []
->                      , docDate = []})
->  [Para [Str "This",Space,Str "is",Space,Str "the",Space,Str "first",
->   Space,Str "paragraph"]
->  ,Para [Str "And",Space,Emph [Str "another"],Str "."]
->  ,BulletList [[Para [Str "item",Space,Str "one"]
->               ,Para [Str "continuation"]]
->              ,[Plain [Str "item",Space,Str "two",Space,Str "and", Space,
->                 Str "a",Space,Link [Str "link"] ("/url","go to url")]]]]
+> myDoc = Pandoc (Meta {unMeta = fromList [("title",
+>           MetaInlines [Str "My",Space,Str "title"])]})
+>         [Para [Str "This",Space,Str "is",Space,Str "the",Space,Str "first",
+>          Space,Str "paragraph"],Para [Str "And",Space,Emph [Str "another"],
+>          Str "."]
+>         ,BulletList [
+>           [Para [Str "item",Space,Str "one"]
+>           ,Para [Str "continuation"]]
+>          ,[Plain [Str "item",Space,Str "two",Space,Str "and",Space,
+>                   Str "a",Space,Link [Str "link"] ("/url","go to url")]]]]
 
 And of course, you can use Haskell to define your own builders:
 
@@ -97,9 +95,10 @@ module Text.Pandoc.Builder ( module Text.Pandoc.Definition
                            , toList
                            , fromList
                            , isNull
-                           -- , Listable(..)
                            -- * Document builders
                            , doc
+                           , ToMetaValue(..)
+                           , HasMeta(..)
                            , setTitle
                            , setAuthors
                            , setDate
@@ -125,6 +124,7 @@ module Text.Pandoc.Builder ( module Text.Pandoc.Definition
                            , link
                            , image
                            , note
+                           , spanWith
                            , trimInlines
                            -- * Block list builders
                            , para
@@ -142,12 +142,14 @@ module Text.Pandoc.Builder ( module Text.Pandoc.Definition
                            , horizontalRule
                            , table
                            , simpleTable
+                           , divWith
                            )
 where
 import Text.Pandoc.Definition
 import Data.String
 import Data.Monoid
 import Data.Maybe (fromMaybe)
+import qualified Data.Map as M
 import Data.Sequence (Seq, (|>), viewr, viewl, ViewR(..), ViewL(..))
 import qualified Data.Sequence as Seq
 import Data.Foldable (Foldable)
@@ -157,9 +159,7 @@ import Data.Data
 import Data.Typeable
 import Data.Traversable
 import Control.Arrow ((***))
-#ifdef GENERICS
 import GHC.Generics (Generic)
-#endif
 
 #if MIN_VERSION_base(4,5,0)
 -- (<>) is defined in Data.Monoid
@@ -175,9 +175,7 @@ infixr 6 <>
 newtype Many a = Many { unMany :: Seq a }
                  deriving (Data, Ord, Eq, Typeable, Foldable, Traversable, Functor, Show, Read)
 
-#ifdef GENERICS
 deriving instance Generic (Many a)
-#endif
 
 toList :: Many a -> [a]
 toList = F.toList
@@ -233,16 +231,51 @@ trimInlines (Many ils) = Many $ Seq.dropWhileL (== Space) $
 -- Document builders
 
 doc :: Blocks -> Pandoc
-doc = Pandoc (Meta [] [] []) . toList
+doc = Pandoc nullMeta . toList
+
+class ToMetaValue a where
+  toMetaValue :: a -> MetaValue
+
+instance ToMetaValue MetaValue where
+  toMetaValue = id
+
+instance ToMetaValue Blocks where
+  toMetaValue = MetaBlocks . toList
+
+instance ToMetaValue Inlines where
+  toMetaValue = MetaInlines . toList
+
+instance ToMetaValue Bool where
+  toMetaValue = MetaBool
+
+instance ToMetaValue a => ToMetaValue [a] where
+  toMetaValue = MetaList . map toMetaValue
+
+instance ToMetaValue a => ToMetaValue (M.Map String a) where
+  toMetaValue = MetaMap . M.map toMetaValue
+
+class HasMeta a where
+  setMeta :: ToMetaValue b => String -> b -> a -> a
+  deleteMeta :: String -> a -> a
+
+instance HasMeta Meta where
+  setMeta key val (Meta ms) = Meta $ M.insert key (toMetaValue val) ms
+  deleteMeta key (Meta ms) = Meta $ M.delete key ms
+
+instance HasMeta Pandoc where
+  setMeta key val (Pandoc (Meta ms) bs) =
+    Pandoc (Meta $ M.insert key (toMetaValue val) ms) bs
+  deleteMeta key (Pandoc (Meta ms) bs) =
+    Pandoc (Meta $ M.delete key ms) bs
 
 setTitle :: Inlines -> Pandoc -> Pandoc
-setTitle t (Pandoc m bs) = Pandoc m{ docTitle = toList t } bs
+setTitle = setMeta "title"
 
 setAuthors :: [Inlines] -> Pandoc -> Pandoc
-setAuthors as (Pandoc m bs) = Pandoc m{ docAuthors = map toList as } bs
+setAuthors = setMeta "author"
 
 setDate :: Inlines -> Pandoc -> Pandoc
-setDate d (Pandoc m bs) = Pandoc m{ docDate = toList d } bs
+setDate = setMeta "date"
 
 -- Inline list builders
 
@@ -315,8 +348,8 @@ math = singleton . Math InlineMath
 displayMath :: String -> Inlines
 displayMath = singleton . Math DisplayMath
 
-rawInline :: Format -> String -> Inlines
-rawInline format = singleton . RawInline format
+rawInline :: String -> String -> Inlines
+rawInline format = singleton . RawInline (Format format)
 
 link :: String  -- ^ URL
      -> String  -- ^ Title
@@ -333,13 +366,18 @@ image url title x = singleton $ Image (toList x) (url, title)
 note :: Blocks -> Inlines
 note = singleton . Note . toList
 
+spanWith :: Attr -> Inlines -> Inlines
+spanWith attr = singleton . Span attr . toList
+
 -- Block list builders
 
 para :: Inlines -> Blocks
 para = singleton . Para . toList
 
 plain :: Inlines -> Blocks
-plain = singleton . Plain . toList
+plain ils = if isNull ils
+               then mempty
+               else singleton . Plain . toList $ ils
 
 -- | A code block with attributes.
 codeBlockWith :: Attr -> String -> Blocks
@@ -349,8 +387,8 @@ codeBlockWith attrs = singleton . CodeBlock attrs
 codeBlock :: String -> Blocks
 codeBlock = codeBlockWith nullAttr
 
-rawBlock :: Format -> String -> Blocks
-rawBlock format = singleton . RawBlock format
+rawBlock :: String -> String -> Blocks
+rawBlock format = singleton . RawBlock (Format format)
 
 blockQuote :: Blocks -> Blocks
 blockQuote = singleton . BlockQuote . toList
@@ -397,6 +435,9 @@ simpleTable :: [Blocks]   -- ^ Headers
 simpleTable headers = table mempty (mapConst defaults headers) headers
   where defaults = (AlignDefault, 0)
 
+divWith :: Attr -> Blocks -> Blocks
+divWith attr = singleton . Div attr . toList
+
 mapConst :: Functor f => b -> f a -> f b
 mapConst = fmap . const
 
